class Config:
    FIREBASE_API_KEY = 'AIzaSyDBbIiCyAFkz_bZQb0hbP5wFB-ioF6xOyw'
    FIREBASE_AUTH_DOMAIN = 'trail-ml-9e15e.firebaseapp.com'
    GQL_ENDPOINT_URL = ''

    PRIMARY_USER_CONFIG_PATH = 'trailconfig.yml'
    SECONDARY_USER_CONFIG_PATH = '~/.config/trail.yml'


DEVELOPMENT_ENDPOINT_URL = 'http://127.0.0.1:5002'
PRODUCTION_ENDPOINT_URL = 'https://trail-ml-9e15e.ew.r.appspot.com'


class ProductionConfig(Config):
    PRESIGNED_URL_ENDPOINT = f'{PRODUCTION_ENDPOINT_URL}/generate_presigned_bucket_url'
    GQL_ENDPOINT_URL = f'{PRODUCTION_ENDPOINT_URL}/graphql'


class DevelopmentConfig(Config):
    PRESIGNED_URL_ENDPOINT = f'{DEVELOPMENT_ENDPOINT_URL}/generate_presigned_bucket_url'
    GQL_ENDPOINT_URL = f'{DEVELOPMENT_ENDPOINT_URL}/graphql'



import os

from trail.libconfig import libconfig
from trail.userconfig.config import MainConfig
from trail.userconfig.cli_configuration_workflow import get_user_credentials, login_user

primary_config_path = os.path.expanduser(libconfig.PRIMARY_USER_CONFIG_PATH)
secondary_config_path = os.path.expanduser(libconfig.SECONDARY_USER_CONFIG_PATH)

if os.getenv('TRAIL_CONFIG'):
    config_path = os.getenv('TRAIL_CONFIG')
elif os.path.isfile(primary_config_path):
    config_path = primary_config_path
elif os.path.isfile(secondary_config_path):
    config_path = secondary_config_path
else:
    [username, password] = get_user_credentials(
        "Configuration file not found. Please provide required information below."
    )
    login_user(username, password, invalid=False, first=True)
    config_path = secondary_config_path

userconfig = MainConfig(config_path)



import yaml


class Config:
    def __init__(self, path, parent, config):
        self.path = path
        self.parent = parent
        self.config = config

    def __getattr__(self, key):
        key = self._snake_to_camel(key)

        try:
            return self.config[key]
        except KeyError:
            raise AttributeError(self._get_missing_option_message(key))

    def save(self):
        if self.parent:
            self.parent.save()
        else:
            with open(self.path, 'w') as f:
                yaml.dump(self.config, f, sort_keys=False)

    def _get_missing_option_message(self, key):
        return f"Configuration option '{key}' required but not found. " \
               f"Refer to README.md to learn more about required options."

    @staticmethod
    def _snake_to_camel(snake):
        parts = snake.split('_')

        return parts[0] + ''.join(i.capitalize() for i in parts[1:])


class MainConfig(Config):
    def __init__(self, config_path):
        with open(config_path, 'r') as f:
            try:
                super().__init__(
                    path=config_path,
                    parent=None,
                    config=yaml.safe_load(f) or {}
                )
            except yaml.YAMLError as e:
                raise RuntimeError(f'Error loading configuration file: {e}')

    def merge(
            self,
            username,
            password,
            project_id,
            parent_experiment_id
    ):
        if username:
            self.config['username'] = username
        if password:
            self.config['password'] = password
            project_config = self.config['projects']
            if project_id:
                project_config['id'] = project_id
            if parent_experiment_id:
                project_config['parentExperimentId'] = parent_experiment_id

    def project(self):
        if 'projects' not in self.config:
            self.config['projects'] = {}

        return ProjectConfig(
            path=self.path,
            parent=self,
            config=self.config['projects']
        )


class ProjectConfig(Config):
    def __init__(self, path, parent, config):
        super().__init__(path, parent, config)

    def update_parent_experiment_id(self, parent_experiment_id):
        self.config['parentExperimentId'] = parent_experiment_id
        self.save()

    def _get_missing_option_message(self, key):
        return (
            f"Configuration option '{key}' in project '{self.alias}' "
            f"required but not found. "
            f"Refer to README.md to learn more about required options."
        )



import os
import yaml
import pyrebase
import requests
from trail.libconfig import libconfig
from gql.transport.exceptions import TransportQueryError
from gql.transport.aiohttp import AIOHTTPTransport
from gql import Client, gql

GET_ALL_PROJECTS = """
    query{
    allProjects {
      id
      title
    }
  }
  """


def create_example_config(username, password, id, parent_experiment_id):
    EXAMPLE_CONFIG = {
        'username': username,
        'password': password,
        'projects': {'id': id, 'parentExperimentId': parent_experiment_id},
    }
    expanded_path = os.path.expanduser(libconfig.SECONDARY_USER_CONFIG_PATH)
    with open(expanded_path, 'w') as f:
        yaml.dump(EXAMPLE_CONFIG, f, sort_keys=False)
    print(
        f"Configuration with given information  was "
        f"automatically created at '{libconfig.SECONDARY_USER_CONFIG_PATH}' \n"
    )


def get_user_credentials(message):
    print(message)
    username = input("Username: ")
    password = input("Password: ")

    print("Validating credentials... \n")
    return [username, password]


def login_user(useremail, userpassword, invalid=False, first=False):
    email = useremail
    password = userpassword
    firebase = pyrebase.initialize_app(
        {
            'apiKey': libconfig.FIREBASE_API_KEY,
            'authDomain': libconfig.FIREBASE_AUTH_DOMAIN,
            'databaseURL': 'THIS_IS_NOT_USED',
            'storageBucket': 'THIS_IS_NOT_USED',
        }
    )
    auth = firebase.auth()
    if invalid:
        selection = input("If you want to retry input r, or to abort input c: ")
        if selection.lower() == 'c':
            exit()
        elif selection.lower() == 'r':
            [new_email, new_password] = get_user_credentials(
                "Please input credentials. \n"
            )
            email = new_email
            password = new_password
        else:
            return login_user(email, password, invalid=True, first=False)
    try:
        user = auth.sign_in_with_email_and_password(email, password)
        token = user['idToken']
        try:
            transport = AIOHTTPTransport(
                libconfig.GQL_ENDPOINT_URL, headers={'authorization': f'Bearer {token}'}
            )
            client = Client(transport=transport)
            result = client.execute(document=gql(GET_ALL_PROJECTS))
            print("Your projects are listed below:\n")
            print("Project ID | Project Title")
            for project in result["allProjects"]:

                print(project['id'] + "     | " + project['title'])

        except TransportQueryError as e:
            print(e)
        project_id = input("Please type in project id: ")
        parent_experiment = input("Please input parent experiment id: ")
        create_example_config(email, password, project_id, parent_experiment)

    except requests.exceptions.HTTPError as e:
        if first:
            TRAIL_LINK = "https://www.trail-ml.com/sign-up"
            print(f"Dont' have an account yet?" f" You can sign up here: {TRAIL_LINK}")
        if e.errno.response.status_code == 400:  # type: ignore
            user = login_user(email, password, invalid=True, first=False)
            return user
        else:
            raise e

    return user['idToken']



import os

from .config import DevelopmentConfig, ProductionConfig


def is_development_environment():
    return os.getenv('TRAIL_ENV') in ['dev', 'development']


if is_development_environment():
    libconfig = DevelopmentConfig
else:
    libconfig = ProductionConfig



import base64
import os.path
import signal
import sys
from collections import defaultdict
from typing import Union

import mlflow
import pyrebase
import requests
from gql import Client, gql
from gql.transport.aiohttp import AIOHTTPTransport
from gql.transport.exceptions import TransportQueryError
from mlflow.entities import Run
from mlflow.tracking import MlflowClient

from .libconfig import libconfig, is_development_environment
from .userconfig import userconfig


class Trail:
    ADD_EXPERIMENT_MUTATION = """
        mutation (
            $projectId: String!,
            $parentExperimentId: String!,
            $title: String!,
            $comments: String!,
            $instanceRunParameters: GenericScalar!,
            $instanceRunMetrics: GenericScalar!,
            $instanceRunComplete: Boolean!
            $hypothesis: String,
            $metricsHistoryEntries: [MetricsHistoryInput],
            $gitCommitHash: String
        ) {
            addExperiment(
                projectId: $projectId,
                parentExperimentId: $parentExperimentId,
                title: $title,
                comments: $comments,
                hypothesis: $hypothesis
                metricsHistory: $metricsHistoryEntries
                gitCommitHash: $gitCommitHash
                instanceRuns: {
                    comment: "",
                    parameters: $instanceRunParameters,
                    metrics: $instanceRunMetrics,
                    isComplete: $instanceRunComplete
                }
            ) {
                experiment {
                    id
                    title
                    comments
                    instanceRuns {
                        id
                        comment
                        parameters
                        metrics
                        isComplete
                    }
                }
            }
        }
    """

    PUT_ARTIFACT_MUTATION = """
        mutation (
            $experimentId: String!,
            $name: String!,
            $base64Data: String!,
            $tags: [String!]
        ) {
            putArtifact(
                experimentId: $experimentId,
                name: $name,
                base64Data: $base64Data,
                tags: $tags,
            ) {
                artifact {
                    id
                    name
                    contentType
                    size
                    tags
                }
            }
        }
    """

    def __init__(
            self,
            experiment_title='Unnamed Run',
            parent_experiment_id=None,
            project_id=None,
            username=None,
            password=None
    ):
        userconfig.merge(
            username=username,
            password=password,
            project_id=project_id,
            parent_experiment_id=parent_experiment_id
        )

        token = self._retrieve_jwt_token(
            email=userconfig.username,
            password=userconfig.password
        )
        transport = AIOHTTPTransport(
            libconfig.GQL_ENDPOINT_URL,
            headers={'authorization': f'Bearer {token}'}
        )
        self.client = Client(transport=transport)
        self.project_config = userconfig.project()
        self.project_id = self.project_config.id
        self.parent_experiment_id = self.project_config.parent_experiment_id
        self.experiment_title = experiment_title
        self.artifacts = []
        self.default_handler = signal.getsignal(signal.SIGINT)
        self.is_complete = True
        self.hypothesis = ""

    def signal_handler(self, signum, frame):
        self.is_complete = False
        print("Latest run is logged.")
        raise KeyboardInterrupt()

    def __enter__(self):
        if mlflow.active_run() is None:
            raise Exception('No active mlflow run found!')
        signal.signal(signal.SIGINT, self.signal_handler)  # type: ignore
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        run = mlflow.active_run()
        if run is None:
            raise Exception('No active mlflow run found!')

        # we fetch the run like this because of
        # https://mlflow.org/docs/latest/python_api/mlflow.html#mlflow.active_run
        materialized_run = mlflow.get_run(run_id=run.info.run_id)

        if materialized_run:
            self._log_experiment(materialized_run)
            self._upload_artifacts(materialized_run)
        signal.signal(signal.SIGINT, self.default_handler)

    def put_artifact(
            self,
            src:  Union[str, bytes],
            name: str,
            tags: Union[str, list, None]
    ):
        """Queues an artifact for upload to Trail.
        The artifact is uploaded when leaving the `with Trail` block.
        The `src` parameter can be either a string or a bytes object. In case
        of a string, it is assumed to be a path to a file. In case of a bytes
        object, it is assumed to be the raw data of the artifact.

        :param src: The artifact path or bytes to upload
        :param name: The name of the artifact
        :param tags: A single tag or a list of tags
        """

        if isinstance(src, str):
            with open(src, 'rb') as f:
                data = f.read()
        elif isinstance(src, bytes):
            data = src
        else:
            raise ValueError('Artifact source must be of type string or bytes')

        if not tags:
            tags = []
        elif isinstance(tags, str):
            tags = [tags]

        self.artifacts.append((data, name, tags,))

    def _upload_artifact(self, data: bytes, name: str, tags: list):
        self.client.execute(
            document=gql(self.PUT_ARTIFACT_MUTATION),
            variable_values={
                'experimentId': self.parent_experiment_id,
                'name': name,
                'base64Data': base64.b64encode(data).decode(),
                'tags': tags
            }
        )

    def put_hypothesis(self, hypothesis: str):
        self.hypothesis = hypothesis

    def _log_experiment(self, run: Run):
        run_id = run.info.run_id
        tags = {
            k: v for k, v in run.data.tags.items()
            if not k.startswith('mlflow.')
        }
        mlflow_artifacts = [
            artifact.path
            for artifact in MlflowClient().list_artifacts(run_id, 'model')
        ]
        d = {  # noqa: F841
            'run_id': run_id,
            'timestamp': run.info.start_time / 1000.0,
            'user': run.info.user_id,
            'artifacts': mlflow_artifacts,
            'tags': tags
        }
        # Convert numeric parameter values to their respective types
        converted_params = run.data.params.copy()
        for key, value in converted_params.items():
            try:
                converted_params[key] = float(value)
            except ValueError:
                pass

        try:
            result = self.client.execute(
                document=gql(self.ADD_EXPERIMENT_MUTATION),
                variable_values={
                    'projectId': self.project_id,
                    'parentExperimentId': self.parent_experiment_id,
                    'title': self.experiment_title,
                    'comments': '',
                    'instanceRunParameters': converted_params,
                    'instanceRunMetrics': run.data.metrics,
                    'instanceRunComplete': self.is_complete,
                    'hypothesis': self.hypothesis,
                    'gitCommitHash': run.data.tags.get("mlflow.source.git.commit", ""),
                    'metricsHistoryEntries': self.get_metric_history_data(
                        run_id=run_id
                    ),
                },
            )

            experiment_id = result['addExperiment']['experiment']['id']
            self.project_config.update_parent_experiment_id(experiment_id)
            self.parent_experiment_id = experiment_id
        except TransportQueryError as e:
            if is_development_environment():
                print(e)

            print(
                'Error uploading experiment data to Trail. Please contact us '
                'if the problem persists.',
                file=sys.stderr
            )

    def _upload_artifacts(self, run: Run):
        run_id = run.info.run_id

        try:
            # upload cached artifacts
            mlflow_artifacts = MlflowClient().list_artifacts(run_id, 'model')
            for artifact in mlflow_artifacts:
                path = artifact.path
                with open(path, 'rb') as f:
                    self._upload_artifact(
                        data=f.read(),
                        name=os.path.basename(path),
                        tags=['mlflow'],
                    )

            for artifact in self.artifacts:
                data, name, tags = artifact
                self._upload_artifact(
                    data=data,
                    name=name,
                    tags=tags,
                )
        except TransportQueryError as e:
            if is_development_environment():
                print(e)

            print(
                'Error uploading experiment artifacts to Trail. Please contact'
                ' us if the problem persists.',
                file=sys.stderr
            )

    def get_metric_history_data(self, run_id):
        client = mlflow.tracking.MlflowClient()
        run = mlflow.active_run()
        mlflowrun = mlflow.get_run(run_id=run.info.run_id)  # type: ignore
        metric_list = list(mlflowrun.data.metrics.keys())
        metrics = defaultdict(list)
        for metric in metric_list:
            metric_history = client.get_metric_history(run_id, metric)
            for entry in metric_history:
                metric_name = entry.key
                data_point = {
                    "value": entry.value,
                    "timeStamp": str(entry.timestamp),
                    "step": entry.step,
                }
                metrics[metric_name].append(data_point)

        metrics_history = []
        for metric in metrics:
            metrics_history.append(
                {"metricName": metric, "history": metrics[metric]}
            )

        return metrics_history

    @staticmethod
    def _retrieve_jwt_token(email, password):
        firebase = pyrebase.initialize_app(
            {
                'apiKey': libconfig.FIREBASE_API_KEY,
                'authDomain': libconfig.FIREBASE_AUTH_DOMAIN,
                'databaseURL': 'THIS_IS_NOT_USED',
                'storageBucket': 'THIS_IS_NOT_USED',
            }
        )
        auth = firebase.auth()

        try:
            user = auth.sign_in_with_email_and_password(email, password)
        except requests.exceptions.HTTPError as e:
            if e.errno.response.status_code == 400:  # type: ignore
                raise Exception('Invalid credentials') from None

            raise e

        return user['idToken']



import os
import requests

from trail.libconfig import libconfig
from trail.trail import Trail
from trail.userconfig import userconfig

cached_token = None


def _retrieve_token() -> str:
    global cached_token

    if cached_token is None:
        cached_token = Trail._retrieve_jwt_token(
            email=userconfig.username, password=userconfig.password
        )

    return cached_token


def _generate_presigned_url(file_path: str, expiration_seconds: int):
    current_working_dir = os.getcwd()
    relative_path = os.path.relpath(file_path, start=current_working_dir)
    project_id = userconfig.project().config["id"]
    destination_blob_name = os.path.join(project_id, relative_path)

    url = (
        f"{libconfig.PRESIGNED_URL_ENDPOINT}?"
        f"destination_blob_name={destination_blob_name}&"
        f"expiration_seconds={expiration_seconds}"
    )
    token = _retrieve_token()
    headers = {"authorization": f"Bearer {token}"}

    try:
        response = requests.get(url, headers=headers)
        if response.status_code == 200:
            data = response.text
            return data
        else:
            return None
    except requests.exceptions.RequestException as e:
        raise e


def upload_file(
    local_file_path: str, expiration_seconds: int, is_absolute_path=False, abs_path=""
) -> None:
    if not is_absolute_path:
        project_directory = os.path.dirname(abs_path)
        local_file_path = os.path.abspath(
            os.path.join(project_directory, local_file_path)
        )

    signed_url = _generate_presigned_url(local_file_path, expiration_seconds)

    if signed_url is None:
        print("You are not authorized to do this action.")
        return

    with open(local_file_path, "rb") as local_file:
        response = requests.put(signed_url, data=local_file)

    if response.status_code == 200:
        print(f"File {local_file_path} uploaded successfully.")
    else:
        print(f"Error uploading file {local_file_path}.")


def upload_folder(local_folder: str, abs_path: str, expiration_seconds=300) -> None:
    project_directory = os.path.dirname(abs_path)
    full_local_folder = os.path.abspath(os.path.join(project_directory, local_folder))
    for root, _, files in os.walk(full_local_folder):
        for file in files:
            local_file_path = os.path.join(root, file)
            upload_file(
                local_file_path,
                is_absolute_path=True,
                expiration_seconds=expiration_seconds,
            )
